MACS_VERSION = "3.0.0a3"
MAX_PAIRNUM = 1000
MAX_LAMBDA  = 100000
FESTEP      = 20
BUFFER_SIZE = 100000                   # np array will increase at step of 1 million items
READ_BUFFER_SIZE = 10000000            # 10M bytes for read buffer size
N_MP = 2                               # Number of processers
